/*
						 ***********
						** Summary **
						 ***********
	(1)	Title:
				Daggerfall Dungeon Scanner

	(2)	Purpose:
			To scan the compiled VRML in Daggerfall dungeon blocks
			(*.rdb) or town blocks (*.rmb), extracting all the required
			polygons and assembling them at their canonical positions
			in 3-space.

	(3)	Non-ANSI features:
*/
#ifdef __GNUC__
#	pragma pack(1)
#	define _GNU_SOURCE
#endif
/*
	(4)	Platform dependencies:
			***  None  ***

	(5)	Change history:
			***  None  ***

	(6)	Input:

			Structures:
				***  None  ***

			Common:
				***  None  ***

			Files:
				arch3d.bsa			--	The object polygon repository.
				blocks.bsa			--	The canonical positions of the
										objects.  *.rdb files describe
										the 187 dungeon variations in
										Daggerfall.  *.rmb files describe
										the 900-odd different town maps.
				texture.???			--	Texture descriptions.

			Assertions about input:
				A general understanding of covariant and contravariant
				tensor algebra is assumed.

				The command line syntax is shown below:

					rooms  [-o] [-i]  room1  room2 ...

				where roomN is the file name of an rdb/rmb file in blocks.bsa.

				-o selects only the object directory.  There are no
				repeated objects, nor is any transformation of coordinates
				performed.

				-i selects both the exterior and interior objects in an
				rmb scenario.  By default, only the building exterior
				is selected.  -i has no effect on rdb scenarios.

				roomN can contain a colon separated list of objects to extract.
				(By default all objects are included)

					rooms s0000000.rdb:74039:74080

				extracts all the 74039 and 74080 objects from the first
				chamber of the Mantellan Crux.

					rooms -i mageaa08.rmb:1

				extracts just the Mages guild hall (from Daggerfall City,
				among other places) but includes all the interior objects
				in it.

				If rooms isn't run from the Daggerfall installation dirctory,
				set TES2 in the environment table to point to the the
				Daggerfall installation to use with the tool.
				
				By default, the extracted objects are arranged in their
				canonical order in the requested block(s).  Which means
				they are properly positioned in 3-space, requiring only
				projection onto the view screen canvas for display.

				In town maps (*.rmb) the unit cell is 4096 by 4096.  The
				standard right hand Cartesian origin is (0, 0, 4096-z)
				before transformation to the right hand system.  The
				homogeneous transformation tensor is given by:

					[e] =  [T][R][x]

				where [e] are the resulting right-hand coordinates,
				[x] is the original coordinate triple from arch3d
				in the homogenized form:
					(x)
					(y)
					(z)
					(1)

				and [T] is
					(1  0  0  Px + Ox       )
					(0  0  1  Oy            )
					(0  1  0  Pz + 4096 - Oz)

				for the interiors of buildings above ground and every
				where else [T] is
					(1  0  0  Px + Ox       )
					(0  0  1  Oy            )
					(0 -1  0  Pz + 4096 - Oz)

				where Ox, Oy, Oz come from Rmb3dObject or RdbPartHeader
				and Px, Pz come from RmbPosition for *.rmb files and
				from mapPItem or mapDItem for *.rdb files.  For a single,
				unpacked rdb file, Px and Pz are taken as 0.  The actual
				value of Px, Pz for an rdb file isn't known until the
				closest packing arrangement of the dungeon using
				the file has been realized.  (See the dungeon thumbnail
				generator in atlas for one way to determine the packing
				order of rdb files in a dungeon)

				Why z has an additional mirror plane for the building
				interiors is a bit surprising.  Presumably it tricks
				the renderer into handling the insides above ground
				the same way as a dungeon.  Perhaps the Bethesda
				tensor handles z by context, rather than total
				geometry.  Negative z also allows them to be clipped
				trivially from the above ground rendering--most buildings
				in the scene will never be entered, and those which are
				have only serial access.

				[R], the homogenized rotation, is constructed from
				Rmb3dObject.extents, and Py from RmbPosition, which
				give the rotation around the x, y, and z axis of the
				3d object to orient it properly in the constructed scene.
				Py is 0 for rdb files.

				For town exteriors, x and z are invariably 0.  For dungeon
				and interior scenes, occasionally x and z are turned.
				Reorientation about y is common in every context.  Note
				that the homogenized rotation matrix is just an ordinary
				rotation matrix with a dihedral column vector adjoined, vis
					(r11 r12 r13  1)
					(r21 r22 r23  1)
					(r31 r32 r33  1)

				Homogeneous coordinates reduce the number of passes through
				the point coordinates that are required to render a scene
				completely.  They can be regarded solely as a performance
				optimization.

	(7)	Output:

			Structures:
				***  None  ***

			Common:
				***  None  ***

			Files:
				***  None  ***

			Assertions about the output:
				It is copious.

	(8)	Public members:

			Title:

			Purpose:

			Parameters:

	(9)	Private members:

			Title:

			Purpose:

			Parameters:

	(10) Subprograms called:
			***  None  ***

	(11) Includes referenced:
			***  None  ***

	(12) Structures declared:
			BsaEntry				--	The stock BSA archive member.
			Handle					--	The memory-mapped 3d object.
			ObjectHeader			--	The preamble to every polygon set.
			PlaneData				--	The face plane data for every polygon
										set.
			PointMarker				--	The collection of points composing a
										a single polygon.
			Query					--	The set of polygons requested from
										a single scenario.
			RdbFixedHeader			--	The preamble to each rdb file.
			RdbFixedTrailer			--	The appendix of the fixed length
										portion of an rdb file.
			Rdb3dObjectDir			--	The rdb record inventory of a 3d
										object rendered in the scene.
			RdbPartAnchor			--	The linked list anchor record.
			RdbPartHeader			--	The fixed header of each element
										in the linked list of items in a
										scene.
			RdbPart0Data			--	The 3d object record extention.
			RdbPart1Data			--	The canonical position of a 3d
										object in the scene.
			RdbPart2Data			--	Unknown.
			RdbPart3Data			--	The canonical position of an
										anciliary texture in the scene.
										(NPCs, foes, lights, etc)
			RmbFlatText				--	Detailed descriptions for certain
										important pseudo-textures.
			RmbFixedHeader			--	The preamble to each rmb file.
			RmbPosition				--	The origin of a building within a
										town map.
			RmbDoorStop				--	The canonical description of a
										building on the town map: faction,
										swank, and type (guild hall, residence,
										etc).
			RmbOffset				--	Working storage for the active
										rendering of the block.
			RmbSize					--	The amount of in-memory storage
										required to describe a town block
										component.
			RmbHeader				--	Header describing the components
										furnishing a single block on the
										town grid.
			Rmb3dObject				--	A single 3D object appearing in
										a town scenario.
			RmbNpc					--	An NPC appearing in a town scenario.
			RmbSection				--	Unknown.
			RmbDoor					--	A doorway in a town scenario.
			RmbFlat					--	A bit of scenery in a twon scenario.


						 ****************
						** Declarations **
						 ****************

					 ***********************
					** Global declarations **
					 ***********************

				  Copyright 2002, Nagual Enterprises
			Please see COPYING for the complete GNU license

					  ********************
					 ** Global functions **
					  ********************
*/

#					include <stdlib.h>
#				ifndef __BORLANDC__
#					include <unistd.h>
#				else
#					define __inline__
					extern void* farmalloc(unsigned long q);
#					include <io.h>
#				endif
#					include <stdio.h>
#					include <string.h>
#					include <sys/types.h>
#					include <sys/stat.h>
#					include <fcntl.h>
#					include <errno.h>
#					include <string.h>
#					include <math.h>

/*
					  *************
					 ** Constants **
					  *************
*/
#				ifdef linux
#					define O_BINARY				0
#					define huge
#					define far
#				endif

#					define FloorMapSize			256
#					define FloorRowSize			16
#					define FloorMapOffset		0x6cb
#					define RmbFixedSize			0x1a78
#					define RmbPositionSize		640
#					define RmbDoorStopSize		832
#					define RmbOffsets			0x5c3
#					define RmbOffsetSize		128
#					define RmbSizes				0x643
#					define RmbSizeSize			128

#					define RdbObjectSize		0x1770
#					define RdbOffsets			0x1784
#					define RdbMaxObjects		0x2ee
#					define RdbOffsetSize		0xbb8
#					define RdbTrailerOffset		0x233c
#					define RdbPartHeaderSize	21
#					define RdbPart1SubdataSize	10
#					define false				0
#					define true					1

/*
					   **********
					  ** Macros **
					   **********
*/

						/*	Answer the start of the face plane data of
							an object from arch3d.bsa */

#					define startPlaneData		\
						(PlaneData*)handleAt(handle, header->toPlaneData)

						/*	Skip to the PointMarker data of a polygon. */

#					define skipData(z)			&((char*)data)[z]

						/*	Answer the start of the point data of a specific
							point in an object from arch3d.bsa */

#					define skipToPoint(z)		\
						(long*)handleAt(handle, (z) + header->toPoints);

						/*	Diagnostic aides. */

#					define print2(style, handle, hd)		\
						print##style(handle, hd);
#					define print3(style, handle, hd, fd)	\
						print##style(handle, hd, fd)

#			ifdef __BORLANDC__
#				define DirSeperator			'\\'
#				define DirSep				"\\"
#				define sortString(x, y)		(stricmp(x, y))
#				define isEqualIs(x, y)		(!stricmp(x, y))
#				define isEqualI(x, y, z)	(!strncmpi(x, y, z))
#				define BitBucket			"nul"
#			else
#				define DirSeperator			'/'
#				define DirSep				"/"
#				define sortString(x, y)		(strcasecmp(x, y))
#				define isEqualIs(x, y)		(!strcasecmp(x, y))
#				define isEqualI(x, y, z)	(!strncasecmp(x, y, z))
#				define BitBucket			"/dev/null"
#			endif

#				define asPoint(e, f)		\
					f[0] = (float)(e[0]); 	\
					f[1] = (float)(e[1]);	\
					f[2] = (float)(e[2]);

#				define asScaledPoint(e, f, s)	\
					f[0] = (float)(e[0]/s);		\
					f[1] = (float)(e[1]/s);		\
					f[2] = (float)(e[2]/s);

#				define asPoints(e, f)		\
					f[0] = (float)(e[0]); 	\
					f[1] = (float)(e[1]);	\
					f[2] = (float)(e[2]);	\
					f[3] = (float)(e[3]);	\
					f[4] = (float)(e[4]);

#				define skip(style)	(&handle[sizeof(style)])

/*
					 **************
					** Structures **
					 **************
*/

					/*	Declare a BSA member element record. */

				struct _BsaEntry
				{
					char name[12];			/*	Member filename part. */
					short fill;				/*	Empty half-word. */
					long fSize;				/*	Size of this archive member. */
				};
				typedef struct _BsaEntry BsaEntry;
#				define newBsaEntryA(z)	((BsaEntry*)malloc((z)*sizeof(BsaEntry)))
#				define newBsaEntry(z)	((BsaEntry*)malloc(z))

					/*	Declare a memory-mapped object. */

				struct _Handle
				{
					long fileSize;
					char huge* object;
				};
				typedef struct _Handle Handle;

					/*	Position the memory stream z at an offset, w. */

#				define handleAt(z, w)		&((z)->object[w])

					/*	Position the arch3d memory stream at an object. */

#				define objectAt(w)			handleAt(object, w)

					/*	Define the polygon query. */

				struct _Query
				{
					char* dungeon;			/*	The scenario d'jure. */
					short* partList;		/*	The subset of parts to use. */
					int filter;				/*	True to filter the parts
												list. */
					int partCount;			/*	The filtered number of parts
												to extract. */
					int handle;				/*	Blocks.bsa handle. */
					int o3dHandle;			/*	Arch3d handle. */
					int objCount;			/*	# of parts in Arch3d. */
					int rmbs;				/*	= 1: want exteriors only.
												= 2: want interiors. */
					long* objDir;			/*	Arch3d directory. */
				};
				typedef struct _Query Query;

					/*	Every 3d object in arch3d has a fixed preamble
						describing the polygons comprising the object. */

				struct _ObjectHeader
				{
					char version[4];		/*	v2.[765] */
					long points;			/*	Vertices in object. */
					long planes;			/*	Face planes in object. */
					unsigned long unknown1;	/*	Garbage. */
					long zip1[2];
					long toSection1;		/*	6 points/plane.
												But usually empty space. */
					long toSection2;		/*	24 bytes/record.  Garbage. */
					long section2Count;		/*	Not always true. */
					unsigned long unknown2;	/*	Garbage. */
					long zip2[2];
					long toPoints;			/*	Offset to the set of points. */
					long toNormals;			/*	Offset to the face normals set. */
					unsigned long unknown3;	/*	Garbage. */
					long toPlaneData;		/*	Offset to plane descriptors. */
				};
				typedef struct _ObjectHeader ObjectHeader;
#				define newObjectHeaderA(z)	\
						(ObjectHeader*)malloc((z)*sizeof(ObjectHeader))

					/*	Every polygon is described by its number of
						vertices and the points in 3-space where those
						vertices reside. */

				struct _PlaneData
				{
					char points;			/*	Vertices in this plane. */
					unsigned char unknown1;
					unsigned short picture;	/*	Texture to paint on this plane. */
					unsigned long unknown2;
				};
				typedef struct _PlaneData PlaneData;

					/*	Not actually used, but documents the structure of
						the plane data dope vector. */

				struct _PointMarker
				{
					long offset;			/*	Offset to point (x,y,z). */
					unsigned long unknown1;	/*	Garbage. */
				};
				typedef struct _PointMarker PointMarker;

					/*	Every dungeon scenario has a fixed preamble. */

				struct _RdbFixedHeader
				{
					unsigned long unknown1;
					unsigned long unknown2;
					unsigned long unknown3;
					unsigned long bodyStart;
					unsigned long unknown5;
				};
				typedef struct _RdbFixedHeader RdbFixedHeader;
#				define newRdbFixedHeaderA(z)	(RdbFixedHeader*)malloc((z)*sizeof(RdbFixedHeader))

					/*	Every dungeon scenario has a fixed appendix. */

				struct _RdbFixedTrailer
				{
					long to3dObjectDir;		/*	Offset to chunks of 0x17,
												forward linked. -2 usually
												terminates the list. */
					unsigned long unknown2;
					unsigned long unknown3;
					unsigned long unknown4;	/*	Sometimes == bodyEnd, but
												usually smaller. */
					unsigned long bodyEnd;
				};
				typedef struct _RdbFixedTrailer RdbFixedTrailer;
#				define newRdbFixedTrailerA(z)	(RdbFixedTrailer*)malloc((z)*sizeof(RdbFixedTrailer))

					/*	There is one 0x17 record for each 3d object part
						number in the object directory.  Occassionally
						scenarios mention objects which are subsequently
						nevered positioned in the scene. */

				struct _Rdb3dObjectDir
				{
					long next;				/*	Offset to next chunk. -2 usually
												terminates the list. */
					unsigned short itemNo;	/*	Item number of this list
												element.  Links run down from
												the max. to 0. */
					unsigned long nextPart;	/*	Always chunk + 0xd */
					unsigned long part[3];	/*	Always {0, -2} */
					unsigned char zip1;
				};
				typedef struct _Rdb3dObjectDir Rdb3dObjectDir;
#				define newRdb3dObjectDirA(z)	(Rdm3dObjectDir*)malloc((z)*sizeof(Rdb3dObjectDir))

				/*	Beginning at bodyStart are at most 4 part directory anchors.

					The linked list relative notation appears to use -1 as an
					empty or never used marker, and -2 as an end-of-list marker.
					That is, empty list head anchors tend to be marked by -1
					while end-of-list/list tail anchors tend to be marked by
					-2.  This usage, however, isn't totally consistent. */

				struct _RdbPartAnchor
				{
					long threads[3];		/*	Anchors an RdbPartHeader
												list. */
					unsigned long unknown1;
				};
				typedef struct _RdbPartAnchor RdbPartAnchor;
#				define newRdbPartAnchorA(z)	(RdbPartAnchor*)malloc((z)*sizeof(RdbPartAnchor))

					/*	The remainder of the rdb file is riddled with relics
						from previous incarnations and trials of the dungeon
						scenario.  Only items appearing in the linked listed
						anchored to a part anchor seem to have any relevance
						to the final version of the scenario.

						But, applying this rule also reveals that some
						scenarios mention objects in their directory which
						are not referenced by the chains of linked lists. */

				/*	Every item in a scene is described with a fixed part header
					followed by data idiosyncratic to the sort of item.  The
					part header links items together and supplies the canonical
					position of the item in Euclidean 3-space. */

				struct _RdbPartHeader
				{
					long nextPart;			/*	Offset to next part header. */
					long priorPart;			/*	Offset to previous header. */
					long e3[3];				/*	Local origin (x, y, z) */
					unsigned char type;		/*	Type of following data. */
					unsigned char data[53];	/*	Accommodates largest case. */
				};
				typedef struct _RdbPartHeader RdbPartHeader;
#				define newRdbPartHeaderA(z)	(RdbPartHeader*)malloc((z)*sizeof(RdbPartHeader))

				/*	Part 3 data describe a 2-d image, sometimes refered to
					as decals.  A decal always faces the camera, no matter
					what position is swiveled to.

					Data1 describes different attributes according to whether
					the decal represents an animate (NPC) or inanimate object.
					For NPCs, it gives the faction number to associate with
					the NPC.  For foes, it may specify the foe type, as shown
					in tpl/foes.src.  For other items, it may specify a spell
					effect to cast when the PC clicks on the item. */

					/*	Encoding types (data3) for data1:

						0x00
						0x01
						0x03
						0x04
						0x05
						0x06
						0x07
						0x09
						0x0b
						0x0c -- Npc prompts and accepts reply (Nulfaga's
								statue, Benefactor, etc)  Data1 codes
								the text.rsc msg to emit, repeated in
								the high two bytes (note parallel deltas):
									0x04 -> 0x151c
									0x17 -> 0x152f
									0x18 -> 0x1530
									0x40 -> 0x1558
								Answer keys for questions are displaced
								by +256 (0x100).  Probably one byte codes
								the prompt, the other the response offset.
						0x0e -- Teleport link.
						0x15 -- Three instances.
						0x16 -- Two instances.
						0x17 -- Usually a flaming brazier on a pedestal.
								Twice a flaming brazier on the ground.
								And a human skull once.  Possibly a
								special sound effect while the PC is
								in the neighborhood.  Low byte values
								match those for 0x1e.
						0x1a -- A treasure pile in n/7.
						0x1b -- Dead bodies that cast lightning when
								clicked.
						0x1d -- A clickable NPC in a dungeon.  High 2
								bytes are the NPC's faction number.
						0x1e -- Make sound when clicked. (the singing
								warrior statue, the young page in Wayrest,
								etc.  Low byte probably indicate which
								noise.
						0x20 -- A quest item marker in n/52.
						0x63 -- A specific foe type.  Spell 99 is tame, and
								may imply these foes have been 'restrained'
								from attacking the PC.  They occur only in
								the main castles, scourg burrow, etc.
						0x64 -- One of the Orcs in Orsinium. */

				struct _RdbPart3Data
				{
					long next;				/*	Offset to data section. */
					unsigned short picture;	/*	A flat texture index. */
					unsigned long data1;	/*	Non-zero for monsters and
												teleports.  Faction # for
												some NPCs; op-codes for
												others. */
					unsigned long link;		/*	Teleporter link. */
					unsigned char data3;	/*	Type of encoding */
				};
				typedef struct _RdbPart3Data RdbPart3Data;
#				define newRdbPart3DataA(z)	(RdbPart3Data*)malloc((z)*sizeof(RdbPart3Data))

				struct _RdbPart2Data
				{
					long next;				/*	Offset to data section. */
					unsigned long data[2];	/*	Type 2 data. */
					short code;
				};
				typedef struct _RdbPart2Data RdbPart2Data;
#				define newRdbPart2DataA(z)	(RdbPart2Data*)malloc((z)*sizeof(RdbPart2Data))

				/*	Part 1 data describe the canonical orientation of one of
					the 3d objects referenced in the dungeon module.  The
					orientation indicates how the object shall be rotated
					in space around its local frame of reference before
					being translated to its canonical position in the scene.

					The rotation is measured in 512 units per pi/2 (90 degree).
					Two-placed objects (levers, platforms) generally have a
					link to a part 0 record to detail the object's motion. */

				struct _RdbPart1Data
				{
					long next;				/*	Offset to data section. */
					long e3[3];				/*	(x,y,z) rotation. */
					short objNo;			/*	Index of this object in the
												object directory. */
					long data1;
					unsigned char unknown1;	/*	Another object #?. */
					long link;				/*	Offset to 10 bytes of
												subdata. */

					/*	Presently unused.  (See RdbPart0Data) */

					unsigned long subdata;	/*	Populated only when link !=
												-2. */
					unsigned char subcode1;
					long sublink;			/*	Generally another type 1 or 3
												record.  When type 3, is often
												a Teleport dungeon marker. */
					unsigned char subcode2;
				};
				typedef struct _RdbPart1Data RdbPart1Data;
#				define newRdbPart1DataA(z)	(RdbPart1Data*)malloc((z)*sizeof(RdbPart1Data))

					/*	Associate descriptive text with a decal number. */

				struct _RmbFlatText
				{
					unsigned short picture;
					char* name;
				};
				typedef struct _RmbFlatText RmbFlatText;

					/*	Every rmb file has a fixed-length preamble. */

				struct _RmbFixedHeader
				{
					unsigned char blocks;
					unsigned char objs;
					unsigned char flats;
				};
				typedef struct _RmbFixedHeader RmbFixedHeader;

					/*	Every rmb file has a table of local origins for all
						the buildings defined in the file. */

				struct _RmbPosition
				{
					long points[5];
				};
				typedef struct _RmbPosition RmbPosition;

					/*	Every rmb file has a table of canonical descriptions
						of the buildings it provides. */

				struct _RmbDoorStop
				{
					short nameHash;			/*	Sign bit on implies the
												remaining bits are a text.rsc
												key number. */
					long name[4];			/*	Always 0 in maps.bsa. */
					short faction;			/*	Faction housed here. */
					long location;			/*	The regular place hash. */
					char place;				/*	Place code. */
					char style;				/*	Shop rating on a 20 point
												scale.  Reduction by 1/4 seems
												to map onto the `rusty relics'
												and `incense' ratings.  The 20
												point scale is almost certainly
												the `bargaining' plea from the
												shopkeeper. */
				};
				typedef struct _RmbDoorStop RmbDoorStop;

				/*	These are the table of machine code address where
					the corresponding blocks of data can be located.
					In the implementation, the machine addresses are
					always half-word bound, so the difference between
					addresses approximates the parallel entry in sizes,
					give or take 1. */
			
				struct _RmbOffset
				{
					unsigned long offset;
				};
				typedef struct _RmbOffset RmbOffset;

				struct _RmbSize
				{
					unsigned long size;
				};
				typedef struct _RmbSize RmbSize;

					/*	The block list header.  links[0] points to the start
						of the block memory arena; links[1] points past the
						end of the arena.  links[2] seems to point to the
						ancestor.  The links are machine address relics. */

				struct _RmbHeader
				{
					unsigned char objs;
					unsigned char flats;
					unsigned char sections;
					unsigned char npcs;
					unsigned char doors;
					unsigned long links[3];
				};
				typedef struct _RmbHeader RmbHeader;

					/*	Every building block in a town scenario has a list
						of 3d objects (from arch3d.bsa) appearing there.
						Usually there is one 3D object for the exterior
						view, and a raft of 3D objects for the interior. */

				struct _Rmb3dObject
				{
					short objClass;
				/*	char objSubclass;
					unsigned char unknown1;*/
					unsigned short objSubclass;
					unsigned long links[3];		/*	Shows a linked list triple
													for the chunk of extra 3D
													objects. */
					unsigned char unknown2[9];	/*	[8] sporadically populated
													in the chunk of extra 3D
													objects.  The values are
													spread continuously over
													the interval [0,255], with
													5 omissions:
														 8 0a 29 44 e9. */
					unsigned long exy[2];		/*  An E3 triple for the */
					unsigned char ez[3];		/*	chunk of extra objects. */
					long e3[3];					/*	Local origin of object. */
					long extents[4];			/*	0-2: (x,y,z) orientation
													of this object. */
					unsigned short unknown3;	/*	Uniformly 0. */
				};
				typedef struct _Rmb3dObject Rmb3dObject;
#				define newRmb3dObjectA(z)	(Rmb3dObject*)malloc((z)*sizeof(Rmb3dObject))

					/*	Each NPC appearing in a town scene is described by
						their position, texture, and faction number.
						Unknown1 is a code describing their function.
						E.g. vends spells, offers quests, etc. */

				struct _RmbNpc
				{
					long e3[3];					/*	The origin for this sprite
													in the scene. */
					unsigned short picture;		/*	The corresponding
													texture. */
					short faction;				/*	The NPC's faction
													number. */
					unsigned char unknown1;		/*	Type of 'service' the NPC
													offers:
													0x01 =
													0x05 = buy/sell
													0x09 =
													0x0d = buy/sell
													0x21 = inn denizen
													0x25 = Childguardable?
													0x29 = prostitutes?
													0x2d =  */
				};
				typedef struct _RmbNpc RmbNpc;
#				define newRmbNpcA(z)		(RmbNpc*)malloc((z)*sizeof(RmbNpc))

				struct _RmbSection
				{
					long e3[3];
					unsigned short unknown1;
					unsigned short unknown2;
				};
				typedef struct _RmbSection RmbSection;
#				define newRmbSectionA(z)	(RmbSection*)malloc((z)*sizeof(RmbSection))

					/*	Each two-placed item (door) has its position
						marked in the scene. */

				struct _RmbDoor
				{
					long e3[3];					/*	Local origin of item. */
					unsigned short unknown1;	/*	Seems to be how wide the
													door opens (512 = pi/2). */
					unsigned short unknown2;	/*	Unknown, but likely */
					unsigned short unknown3;	/*	indicates whether the */
					unsigned char unknown4;		/*	door is locked at sunset/
													sunrise v unlocked and
													how strong the lock is. */
				};
				typedef struct _RmbDoor RmbDoor;
#				define newRmbDoorA(z)		(RmbDoor*)malloc((z)*sizeof(RmbDoor))

					/*	Every inanimate object in a scene has its
						canonical position and texture. */

				struct _RmbFlat
				{
					long e3[3];					/*	Local origin of item. */
					unsigned short picture;		/*	Item's texture. */
					unsigned short unknown1;	/*	May refer to animation speed. */
					unsigned char unknown2;		/*	A type class marker. */
				};
				typedef struct _RmbFlat RmbFlat;
#				define newRmbFlatA(z)		(RmbFlat*)malloc((z)*sizeof(RmbFlat))

/*					   **********
					  ** Common **
					   **********
*/

					double degreesToRadians = M_PI / 180.0;
					double twoPi = 2.0 * M_PI;
					double mTwoPi;
					double rotor[9];
					double scale = 512.0 / 90.0;
					char* o3d = NULL;
					char typeSizes[] = {
						'\0', '\x1c', '\x0f', '\x10', '\0', '\x33', '\0', '\0'};

					RmbFlatText flatNames[] = {
						/*{0x2f80, "Dungeon exit"},*/
						{0x3200, "More death"},
						{0x5f00, "Still lifes"},
						{0x6382, "Daedra"},
						{0x6383, "Summon"},
						{0x6384, "?Resc?"},
						{0x6385, "Court"},
						{0x6386, "?Pic?"},
						{0x6387, "Effect"},
						{0x6388, "Enter"},
						{0x6389, "Path"},
						{0x638a, "Start"},
						{0x638b, "Quest"},
						{0x638c, "Create"},
						{0x638d, "Prison"},
						{0x638e, "Prison exit"},
						{0x638f, "Random monster"},
						{0x6390, "Monster"},
						{0x6391, "Teleport"},
						{0x6392, "Quest item"},
						{0x6393, "Random treasure"},
						{0x6394, "Random flat"},
						{0x6395, "Upstairs"},
						{0x6396, "Downstairs"},
						{0x6400, "House furnishing"},
						{0x6480, "Animals"},
						{0x6500, "Art"},
						{0x6580, "Pillars"},
						{0x6600, "Clothing"},
						{0x6680, "Containers"},
						{0x6700, "Death"},
						{0x6780, "Equipment"},
						{0x6880, "Library"},
						{0x6900, "Lights"},
						{0x6980, "Miscellany"},
						{0x6a00, "Outdoors"},
						{0x6a80, "Plants"},
						{0x6b00, "Tools"},
						{0x6c80, "Treasure"},
						{0x6d00, "Pots and pans"},
						{0x9600, "Cave formations"},
						{0xffff, NULL},
					};

					long thisPart;
					int elements;
					Query theQuery = {NULL, NULL, 0};
					RdbFixedTrailer* trailer;
					char textureDesc[32];
					char* theData = NULL;
					char theFlats[256];
					char* flats;

/*
				 **********************
				** Local declarations **
				 **********************

					  *************
					 ** Variables **
					  *************

					***  None  ***

						 ********
						** Text **
						 ********


							 *******************
							** Private methods **
							 *******************
*/

		/*	Borland C lseek() is b0rken for file sizes in the
			range 32k-64k, which is where the mapnames sizes
			happen to fall.  Redefine lseek to work right.
			Note that the compiler will crab the method should
			return a value, and it does answer the long result
			in AX:DX, but the compiler has no extension to
			indicate this requirement. (Using the obvious
			`return _AX' apparently cues it that it may freely
			trash DX, and it happily does so)  Note also
			that the coding is tuned for the compact memory
			model. */

#ifdef __BORLANDC__
#	define _asm asm
long Seek
(	int handle,
	long off,
	int how)
{
	_asm {
		lea di,off
		mov dx,ss:[di]
		mov cx,ss:[di+2]
		mov ax,how
		mov ah,42h
		mov bx,handle
		int 21h
	}
/*	return;*/
}
#else
#	define Seek(x,y,z)		lseek(x, y, z)
#endif

long* buildObjDir
(	int handle,
	char* path,
	int* dirCount)
{
		/*	Build an unsorted directory of objects. */

	short dirNo;
	long fileSize;
	long dirStart;
	unsigned int rsize;
	long dirSize;
	long* keys;
	char huge* dPtr;

	read(handle, &dirNo, sizeof(dirNo));
	dirSize = (long)dirNo << 3;
	*dirCount = dirNo;
	fileSize = Seek(handle, 0L, SEEK_END);
	dirStart = fileSize - dirSize;

	printf("#-- %hd directory entries in %s.\n", dirNo, path);
	printf("#-- Scrolling to start of directory at %#lx.\n", dirStart);
	if (Seek(handle, dirStart, SEEK_SET) != dirStart)
	{
		fprintf(stderr, "Can't scroll to arch3d.bsa directory.\n");
		exit(1);
	}
#ifdef __BORLANDC__
	keys = (long*)farmalloc(dirSize);
	if (!keys)
	{
		fprintf(stderr, "Not enough memory for directory.\n");
		exit(1);
	}
	dPtr = (char huge*)keys;
fprintf(stderr, "Reading first batch from %d.\n", handle);
	if ((rsize = _read(handle, dPtr, 0xfffc)) != 0xfffc)
	{
		fprintf(stderr, "1 Read only %#x bytes (%d).\n", rsize, errno);
		exit(1);
	}
	dirSize -= 65532L;
	dPtr = &dPtr[65532U];
	if ((rsize = _read(handle, dPtr, dirSize)) != dirSize)
	{
		fprintf(stderr, "3 Read only %#x bytes (%d).\n", rsize, errno);
		exit(1);
	}
#else
	keys = (long*)malloc(dirSize);
	if ((rsize = read(handle, keys, dirSize)) != dirSize)
	{
		fprintf(stderr, "Read only %#x bytes (%d).\n", rsize, errno);
		exit(1);
	}
#endif
	return keys;
}

long openScenario
(	Query* query)
{
		/*	Answer the size of the requested scenario in blocks.bsa.
			Scroll the blocks file handle to the start of the scenario. */

	long dirLen;
	long seed = 4L;
	short count = 0;

	/*	Scan the archive directory for the map files associated with
		the present query.  The member directory is appended
		at the end of the blocks.bsa file, so scroll to that point to
		begin searching for scenario names. */

	Seek(query->handle, 0L, SEEK_SET);
	read(query->handle, &count, sizeof(count));
	dirLen = Seek(query->handle, 0L, SEEK_END);
	dirLen -= (long)(18*count);
	dirLen = Seek(query->handle, dirLen, SEEK_SET);
	printf("#-- Scanning %hd entries for %s.\n", count, query->dungeon);

	/*	Inject the seed offset until the scenario is located. */

	while (count--)
	{
		BsaEntry me;

		read(query->handle, me.name, sizeof(BsaEntry));
		if (isEqualI(me.name, query->dungeon, 12))
		{
			printf("#-- Found %s at %08lx.\n", query->dungeon, seed);
			Seek(query->handle, seed, SEEK_SET);
			return me.fSize;
		}
		seed += me.fSize;
	}
	printf("Unable to locate scenario %s in blocks.bsa.\n", query->dungeon);
	return 0L;
}

Handle* findObject
(	Query* query,
	Handle* object,
	long key)
{
		/*	Answer the file offset to the beginning of the specified
			3d object. */

	long seed = 4L;
	int count = query->objCount;
	long huge* directory = query->objDir;
	int handle = query->o3dHandle;

	while (count--)
		if (*directory++ == key)
		{
#		ifdef __BORLANDC__
			long size = *directory;

			object->fileSize = *directory;
			object->object = (char huge*)farmalloc(*directory);
			Seek(handle, seed, SEEK_SET);
			_read(handle, object->object, size > 65532L ? 0xfffc : size);
			size -= 65532L;
			if (size > 0)
				_read(handle, &object->object[65532U], size);
#		else
			object->fileSize = *directory;
			object->object = (char*)malloc(*directory);
			Seek(handle, seed, SEEK_SET);
			read(handle, object->object, object->fileSize);
#		endif
			return object;
		}
		else
			seed += *directory++;
	return NULL;
}

__inline__ double dot
(	double* a3dPoint1,
	double* a3dPoint2)
{
		/*	Return the dot (inner) product of two vectors. */

	return	(a3dPoint1[0] * a3dPoint2[0]) +
			(a3dPoint1[1] * a3dPoint2[1]) +
			(a3dPoint1[2] * a3dPoint2[2]);
}

double reduce2Pi
(	double point)
{
		/*	Reduce the range of a point to [0, 2pi]; */

	if (point < 0.0)
	{
		while (point < 0.0)
			point += twoPi;
		return point;
	}
	while (point > twoPi)
		point -= twoPi;
	return point;
}

double reduce360
(	double point)
{
		/*	Reduce the range of a point to [0, 2pi]. */

	if (point < 0.0)
	{
		while (point < 0.0)
			point += 360.0;
		return point;
	}
	while (point > 360.0)
		point -= 360.0;
	return point;
}

void _rotate
(	double* self,
	double* cosines)
{
		/*	If cosines are the collection of direction cosines, transform
			the receiver from its original basis to the one indicated by
			those cosines.  The receiver becomes the rotated point. */

	double x = dot(self, &cosines[0]);
	double y = dot(self, &cosines[3]);
	double z = dot(self, &cosines[6]);
	self[0] = x;
	self[1] = y;
	self[2] = z;
	return;
}

void rotateX
(	double degrees,
	double* rotor)
{
		/*	Answer the set of direction cosines to rotate X by degrees. */

	double rad = degrees * degreesToRadians;
	double _sin = sin(rad);
	double _cos = cos(rad);
	rotor[0] = 1.0;
	rotor[1] = 0.0;
	rotor[2] = 0.0;
	rotor[3] = 0.0;
	rotor[4] = _cos;
	rotor[5] = _sin;
	rotor[6] = 0.0;
	rotor[7] = -_sin;
	rotor[8] = _cos;
	return;
}

void rotateY
(	double degrees,
	double* rotor)
{
		/*	Answer the set of direction cosines to rotate Y by degrees. */

	double rad = degrees * degreesToRadians;
	double _sin = sin(rad);
	double _cos = cos(rad);
	rotor[0] = _cos;
	rotor[1] = 0.0;
	rotor[2] = -_sin;
	rotor[3] = 0.0;
	rotor[4] = 1.0;
	rotor[5] = 0.0;
	rotor[6] = _sin;
	rotor[7] = 0.0;
	rotor[8] = _cos;
	return;
}

void rotateXYZ
(	double degree1,
	double degree2,
	double degree3,
	double* rotor)
{
		/*	Answer the rotation matrix to turn degree1 in x, degree2 in y, and
			degree3 in z.  This is just the result of multiplying rotX, rotY
			amd rotZ together by hand to take advantage of the elemental
			reduction afforded from known zero quantities. */

	double rad1 = degree1 * degreesToRadians;
	double rad2 = degree2 * degreesToRadians;
	double rad3 = degree3 * degreesToRadians;
	double _sin1 = sin(rad1);
	double _cos1 = cos(rad1);
	double _sin2 = sin(rad2);
	double _cos2 = cos(rad2);
	double _sin3 = sin(rad3);
	double _cos3 = cos(rad3);
	rotor[0] = (_cos2 * _cos3);
	rotor[1] = (_cos2 * _sin3);
	rotor[2] = -_sin2;
	rotor[3] = ((_cos1 * -_sin3) + (_sin1 * _sin2 * _cos3));
	rotor[4] = ((_cos1 * _cos3) + (_sin1 * _sin2 * _sin3));
	rotor[5] = (_sin1 * _cos2);
	rotor[6] = ((_sin1 * _sin3) + (_cos1 * _sin2 * _cos3));
	rotor[7] = ((-_sin1 * _cos3) + (_cos1 * _sin2 * _sin3));
	rotor[8] = (_cos1 * _cos2);
	return;
}

int matchObject
(	Query* query,
	unsigned short obj)
{
		/*	Answer true if the specified object bas been requested by
			the user. */

	int count = query->partCount;
	short* parts = query->partList;

	while (count--)
	{
		if (obj < *parts)
			break;
		if (*parts++ == obj)
			return 1;
	}
	return 0;
}

char* describeTexture
(	char* endPath,
	char* path,
	int texture)
{
		/*	Answer the texture description for the specified texture file. */

	int handle;
	char file[32];

	sprintf(file, "arena2/texture.%03d", texture);
	endPath = stpcpy(endPath, file);
	handle = open(path, O_RDONLY|O_BINARY);
	if (handle < 0)
	{
		printf("-- Can't open texture file '%s'\n", path);
		return "";
	}
	lseek(handle, 2L, SEEK_SET);
	*textureDesc = '\0';
	read(handle, textureDesc, 26);
	close(handle);
	if (*textureDesc == '\t')
		*textureDesc = ' ';
	return textureDesc;
}

char* findFlat
(	unsigned short picture)
{
		/*	Translate the known flat picture names.  Special
			case the editor (invisible) flat names to give
			specific descriptions instead of the stock texture
			file description. */

	RmbFlatText* flatns = flatNames;
	while (flatns->name)
	{
		unsigned short code = flatns->picture & 0xff80;

		if ((picture & 0xff80) == code)
		{
			if (code != 0x6380)
				return flatns->name;
			if (flatns->picture == picture)
				return flatns->name;
		}
		flatns++;
		if (picture < flatns->picture)
			break;
	}
	return describeTexture(flats, theFlats, ((unsigned int)picture)>>7);
}

printPosition
(	char* handle)
{
		/*	Describe the position of a block on the town map. */

	RmbPosition* flat = (RmbPosition*)handle;
	double e5[5];

	asPoints(flat->points, e5)
	/*printf(	"%08x %08x %08x %08x %08x\n",*/
	printf(	"%11.4f %11.4f %11.4f %11.4f %11.4f\n",
			e5[0],
			e5[1],
			e5[2],
			e5[3],
			e5[4]);
	return 1;
}

int printDoorStop
(	char* handle)
{
		/*	Describe a building on the town map. */

	RmbDoorStop* flat = (RmbDoorStop*)handle;
	printf(	"%04x %08x %08x %08x %08x %4d %08x %d %d\n",
			flat->nameHash,
			flat->name[0],
			flat->name[1],
			flat->name[2],
			flat->name[3],
			flat->faction,
			flat->location,
			(int)flat->place,
			(int)flat->style);
	return 1;
}

printOffset
(	char* handle)
{
	RmbOffset* flat = (RmbOffset*)handle;
	printf(	"%08x\n",
			flat->offset);
	return 1;
}

printSize
(	char* handle)
{
	RmbSize* flat = (RmbSize*)handle;
	printf(	"%08x\n",
			flat->size);
	return 1;
}

int print3dObject
(	char* handle)
{
		/*	Describe a 3d object on the town map. */

	Rmb3dObject* flat = (Rmb3dObject*)handle;
	long key;
	char code;
	long sub;
	double e3[3];
	double f3[3];

	sub = flat->objSubclass & 0xff;
	code = sub >= 100 ? '*' : ' ';
	key = sub + ((long)flat->objClass * 100);
	asPoint(flat->e3, e3)
	asPoint(flat->extents, f3)
	printf(	"%c%05x %04hx %04hx %08x %08x %08x %08x %08x %02x ",
			code,
			key,
			flat->objClass,
			flat->objSubclass,
			flat->links[0],
			flat->links[1],
			flat->links[2],
			flat->unknown2[0],
			flat->unknown2[4],
			*(unsigned int*)&flat->unknown2[8]&0xff);
	printf(	"%08x %08x %06x %11.4f %11.4f %11.4f %08x  %11.4f %08x %08x %04x\n",
			flat->exy[0],
			flat->exy[1],
			*((unsigned int*)flat->ez)&0xffffff,
			e3[0],
			e3[2],
			-e3[1],
			flat->extents[0],
			f3[1],
			flat->extents[2],
			flat->extents[3],
			flat->unknown3);
	return 1;
}

int printFlat
(	char* handle)
{
		/*	Describe a decal on the town map. */

	RmbFlat* flat = (RmbFlat*)handle;
	double e3[3];

	asPoint(flat->e3, e3)
	printf(	"%11.4f %11.4f %11.4f 0x%04hx 0x%04hx 0x%02x (%s)\n",
			e3[0],
			e3[2],
			-e3[1],
			flat->picture,
			flat->unknown1,
			(unsigned int)flat->unknown2,
			findFlat(flat->picture));
	return 1;
}

int printFlatLabel
(	char* handle,
	long* origin)
{
		/*	Describe a decal on the town map. */

	RmbFlat* flat = (RmbFlat*)handle;
	double e3[3];
	long label[3];
	char* flatText = findFlat(flat->picture);

	label[0] = origin[0] + flat->e3[0];
	label[1] = origin[1] + flat->e3[1];
	label[2] = origin[2] + flat->e3[2];
	asPoint(label, e3)
	printf(	"# %s: %d\n#set label '%c' at %12.6f, %12.6f, %12.6f\n",
			flatText,
			(int)flat->picture&0x7f,
			*flatText,
			e3[0],
			e3[2],
			-e3[1]);
	return 1;
}

int printNpc
(	char* handle)
{
		/*	Decribe an NPC on the town map. */

	RmbNpc* flat = (RmbNpc*)handle;
	double e3[3];

	asPoint(flat->e3, e3)
	printf(	"%11.4f %11.4f %11.4f 0x%04hx 0x%04hx 0x%02x (%s)\n",
			e3[0],
			e3[2],
			-e3[1],
			flat->picture,
			flat->faction,
			(unsigned int)flat->unknown1,
			findFlat(flat->picture));
	return 1;
}

int printSection
(	char* handle)
{
	RmbSection* flat = (RmbSection*)handle;
	double e3[3];

	asPoint(flat->e3, e3);
	printf(	"%11.4f %11.4f %11.4f 0x%04hx 0x%04hx\n",
			e3[0],
			e3[2],
			-e3[1],
			flat->unknown1,
			flat->unknown2);
	return 1;
}

int printDoor
(	char* handle)
{
		/*	Decribe a doorway position on the town map. */

	RmbDoor* flat = (RmbDoor*)handle;
	double e3[3];

	asPoint(flat->e3, e3);
	printf(	"%11.4f %11.4f %11.4f 0x%04hx 0x%04hx 0x%04hx 0x%02x\n",
			e3[0],
			e3[2],
			-e3[1],
			flat->unknown1,
			flat->unknown2,
			flat->unknown3,
			flat->unknown4);
	return 1;
}

/*
							 ******************
							** Public methods **
							 ******************
*/

char* printRmbObjects
(	char* handle,
	RmbHeader* header)
{
		/*	Describe the 3d objects on a town map. */

	int count = header->objs;

	if (count)
		printf("--  3D object: %d at 0x%x\n", count, handle - theData);
	while (count--)
	{
		if (!print3dObject(handle))
			return NULL;
		handle = skip(Rmb3dObject);
	}
	return handle;
}

char* printDoors
(	char* handle,
	RmbHeader* header)
{
		/*	Describe the doorway positions on the town may. */

	int count = header->doors;

	if (count)
		printf("-- Doors: %d\n", count);
	while (count--)
	{
		if (!printDoor(handle))
			return NULL;
		handle = skip(RmbDoor);
	}
	return handle;
}

char* printDoorStops
(	char* handle,
	RmbFixedHeader* fixed)
{
		/*	Describe the buildings on the town map. */

	int count = RmbDoorStopSize / sizeof(RmbDoorStop);
	int count1 = fixed->blocks;

	if (count1)
		printf("-- Doorstops: %d\n", count1);
	while (count--)
	{
		if (count1-- > 0)
		{
			if (!printDoorStop(handle))
				return NULL;
		}
		handle = skip(RmbDoorStop);
	}
	return handle;
}

char* printFlats
(	char* handle,
	RmbHeader* header)
{
		/*	Describe the decorations in a town scene. */

	int count = header->flats;

	if (count)
		printf("-- Flats: %d\n", count);
	while (count--)
	{
		if (!printFlat(handle))
			return NULL;
		handle = skip(RmbFlat);
	}
	return handle;
}

char* printFlatLabels
(	char* handle,
	int count,
	long* origin)
{
		/*	Describe the decorations in a town scene. */

	if (count)
		printf("#-- Flats: %d\n", count);
	while (count--)
	{
		if (!printFlatLabel(handle, origin))
			return NULL;
		handle = skip(RmbFlat);
	}
	return handle;
}

int printFloorMap
(	int handle,
	char* rmbFile)
{
	char blocks[FloorMapSize];
	char* blk;
	int rsize;
	int count = FloorRowSize;

	if (lseek(handle, FloorMapOffset, SEEK_SET) != FloorMapOffset)
	{
		fprintf(stderr, "Can't scroll to directory in %s.\n", rmbFile);
		return 0;
	}
	errno = 0;
	memset(blocks, 0, FloorMapSize);
	if ((rsize = read(handle, (void*)blocks, FloorMapSize)) != FloorMapSize)
	{
		fprintf(stderr, "Read only %#x bytes in %s (%d).\n", rsize, rmbFile, errno);
		return 0;
	}
	printf("%s:\n", rmbFile);
	blk = &blocks[FloorRowSize-1];
	while (count--)
	{
		int col = FloorRowSize;
		char* bPtr = blk;
		while (col--)
		{
			printf("%02x", ((unsigned int)*--bPtr)&0xff);
		}
		printf("\n");
		blk = &blk[FloorRowSize];
	}
}

char* printNpcs
(	char* handle,
	RmbHeader* header)
{
		/*	Describe the NPCs in a town scene. */

	int count = header->npcs;

	if (count)
		printf("-- Npcs: %d\n", count);
	while (count--)
	{
		if (!printNpc(handle))
			return NULL;
		handle = skip(RmbNpc);
	}
	return handle;
}

char* printOffsets
(	char* handle,
	RmbFixedHeader* fixed)
{
	int count = RmbOffsetSize / sizeof(RmbOffset);
	int count1 = fixed->blocks;

	if (count1)
		printf("-- Offsets: %d\n", count1);
	while (count--)
	{
		if (count1-- > 0)
			printOffset(handle);
		handle = skip(RmbOffset);
	}
	return handle;
}

char* printPositions
(	char* handle,
	RmbFixedHeader* fixed)
{
	int count = RmbPositionSize / sizeof(RmbPosition);
	int count1 = fixed->blocks;

	if (count1)
		printf("-- Positions: %d\n", count1);
	while (count--)
	{
		if (count1-- > 0)
			printPosition(handle);
		handle = skip(RmbPosition);
	}
	return handle;
}

char* printSections
(	char* handle,
	RmbHeader* header,
	RmbFixedHeader* fixed)
{
	int count = header->sections;

	if (count)
		printf("-- Sections: %d\n", count);
	while (count--)
	{
		printSection(handle);
		handle = skip(RmbSection);
	}
	return handle;
}

int printObjectHeader
(	ObjectHeader* header,
	char* name)
{
		/*	Dump the header contents on stdout. */

	printf(	"%5.5s:%4ld%4ld %08lx %08lx %08lx\n",
			name,
			header->points,
			header->planes,
			header->unknown1,
			header->unknown2,
			header->unknown3);
	return 1;
}

char* printSizes
(	char* handle,
	RmbFixedHeader* fixed)
{
	int count = RmbSizeSize / sizeof(RmbSize);
	int count1 = fixed->blocks;

	if (count1)
		printf("-- Sizes: %d\n", count1);
	while (count--)
	{
		if (count1-- > 0)
			printSize(handle);
		handle = skip(RmbSize);
	}
	return handle;
}

int printPlaneData
(	int handle,
	ObjectHeader* header,
	char* name)
{
		/*	Dump the dope vectors of the polygons in this object. */

	int count = header->planes;

	Seek(handle, header->toPlaneData, SEEK_SET);
	while (count--)
	{
		PlaneData data;
		int points;
		int col = 0;
		int size;
		unsigned long* point;
		unsigned long* pointList;

		read(handle, &data, sizeof(PlaneData));
		size = (int)data.points << 3;
		point = pointList = (unsigned long*)malloc(size);
		read(handle, point, size);
		printf(	"%5.5s: %ld %04hx %02lx %08lx\n",
				name,
				data.points,
				data.picture,
				(unsigned long)data.unknown1,
				data.unknown2);
		points = data.points;
		while (points--)
		{
			if (!(col & 3))
				printf("      ");
			printf(	" %3ld %08lx",
					point[0]/12,
					point[1]);
			point = &point[2];
			if (!(++col & 3))
				printf("\n");
		}
		if (col & 3)
			printf("\n");
		printf("\n");
		free(pointList);
	}
	return 1;
}

int gnuPlanes
(	Handle* handle,
	char* name,
	long* translate,
	long* rotate,
	int rectify)
{
		/*	Write the plane points in gnuplot format, translating
			them into their canonical position inside the room frame. */

	ObjectHeader* header = (ObjectHeader*)handle->object;
	int count = header->planes;
	PlaneData* data;
	double t3[3];
	double r3[3];

	data = startPlaneData;
	asPoint(translate, t3);
	printf(	"#trn  %12.6f %12.6f %12.6f\n",
			t3[0],
			t3[2],
			-t3[1]);

	/*	By direct observation, a rotation of 512 units corresponds to
		a 90 degree clockwise turn in the Mantellan Crux.  Which implies
		a ceiling of 4096 units.  However the observed span of units
		in the *rdb files occassionally reachs double that.  Most settings
		only rotate around the y axis.

		For output, rotate the ensemble pi radians around the x axis to
		rectify the image for visual inspection. */

	r3[0] = reduce360((float)(rotate[0]) / scale);
	r3[1] = reduce360((float)(rotate[1]) / scale);
	r3[2] = reduce360((float)(rotate[2]) / scale);
	printf(	"#rot  %12.6f %12.6f %12.6f\n",
			r3[0],
			r3[2],
			-r3[1]);
/*	rotateY(r3[1], rotor);*/
	rotateXYZ(r3[0], r3[1], r3[2], rotor);
	while (count--)
	{
		int points;
		unsigned long* point;
		double firstPoint[3];
		double avg[3] = {0.0, 0.0, 0.0};
		double f3[3];
		double tScale;

		point = (unsigned long*)skipData(sizeof(PlaneData));
		points = data->points;
		tScale = (float)(points<<8);
		while (points)
		{
			long* pint;
			double e3[3];

			/*	Convert the fixed decimal value to floating point. */

			pint = skipToPoint(point[0]);
			asScaledPoint(pint, e3, 256.0);

			/*	Rotate, then translate the point to its canonical
				position in space. */

			_rotate(e3, rotor);
			e3[0] += t3[0];
			e3[1] += t3[1];
			e3[2] += t3[2];
			printf(	" %12.6f %12.6f %12.6f\n",
					(rectify ? e3[0] : e3[0]),
					(rectify ? e3[2] : e3[2]),
					(rectify ?-e3[1] : e3[1]));

			/*	Save a copy of the first point in the polygon. */

			if (points == data->points)
				memcpy(firstPoint, e3, sizeof(e3));
			avg[0] += pint[0];
			avg[1] += pint[1];
			avg[2] += pint[2];
			point = &point[2];
			points--;
		}
		printf(	" %12.6f %12.6f %12.6f\n",
				(rectify ? firstPoint[0] : firstPoint[0]),
				(rectify ? firstPoint[2] : firstPoint[2]),
				(rectify ?-firstPoint[1] : firstPoint[1]));

		/*	Fabricate a diagnostic label to render for this object. */

		asScaledPoint(avg, f3, tScale);
		f3[0] += t3[0];
		f3[1] += t3[1];
		f3[2] += t3[2];
		if (!count)
			printf(	"#set label '%s' at  %12.6f, %12.6f, %12.6f\n",
					name,
					t3[0],
					t3[2],
					-t3[1]);
		printf("\n\n");

		/*	Advance to the next polygon in the object. */

		data = (PlaneData*)point;
	}
	return 1;
}

int printPointsAt
(	int handle,
	long offset,
	int count,
	char* name)
{
		/*	Print the points comprising one polygon. */

	int index = 0;
	long eMin[3] = {0, 0, 0};
	long eMax[3] = {0, 0, 0};
	double e3[3];

	Seek(handle, offset, SEEK_SET);
	while (count--)
	{
		long point[3];

		read(handle, point, sizeof(point));
		asScaledPoint(point, e3, 256.0);
		printf(	"[%d] %12.6f %12.6f %12.6f\n",
				index++,
				e3[0],
				e3[1],
				e3[2]);
		if (point[0] < eMin[0])
			eMin[0] = point[0];
		if (point[1] < eMin[1])
			eMin[1] = point[1];
		if (point[2] < eMin[2])
			eMin[2] = point[2];
		if (point[0] > eMax[0])
			eMax[0] = point[0];
		if (point[1] > eMax[1])
			eMax[1] = point[1];
		if (point[2] > eMax[2])
			eMax[2] = point[2];
	}
	asScaledPoint(eMin, e3, 256.0);
	printf(	"-- Range: %12.6f %12.6f %12.6f ",
			e3[0],
			e3[1],
			e3[2]);
	asScaledPoint(eMax, e3, 256.0);
	printf(	"to %12.6f %12.6f %12.6f\n",
			e3[0],
			e3[1],
			e3[2]);
}

int close3dObject
(	Handle* handle)
{
		/*	Relinquish a 3d object handle. */

	free(handle->object);
	handle->fileSize = 0;
	handle->object = NULL;
	free(handle);
	return 1;
}

Handle* open3dObject
(	Query* query,
	char* objName)
{
		/*	Memory map the requested 3d object and return
			a handle to it. */

	Handle* object;
	long objNo;

	object = (Handle*)malloc(sizeof(Handle));
	objNo = atol(objName);
	if (!findObject(query, object, objNo))
	{
		fprintf(stderr, "Can't find object %d in arch3d.bsa.\n", objNo);
		return NULL;
	}
	return object;
}

Handle* open3dObjectByNo
(	Query* query,
	long objNo)
{
		/*	Memory map the requested 3d object and return
			a handle to it. */

	Handle* object;

	object = (Handle*)malloc(sizeof(Handle));
	if (!findObject(query, object, objNo))
	{
		fprintf(stderr, "Can't find object %d in arch3d.bsa.\n", objNo);
		return NULL;
	}
	return object;
}

int printTownObject
(	Query* query,
	char* handle,
	int* count,
	long* origin,
	int rectify)
{
		/*	Describe the polygons comprising an object in town. */

	Rmb3dObject* flat = (Rmb3dObject*)handle;
	long key;
	char code;
	long sub;
	Handle* object;
	int newCount;
	char part3d[16];
	long localOrigin[3];
	long localOrient[3];

	sub = flat->objSubclass & 0xff;
	code = sub >= 100 ? '*' : ' ';
	key = sub + ((long)flat->objClass * 100);
	object = open3dObjectByNo(query, key);
	if (!object)
	{
		printf("Can't open 3d object %s\n", part3d);
		exit(1);
	}
	newCount = *count + ((ObjectHeader*)object->object)->planes - 1;
	sprintf(part3d, "%d", key);
	printf(	"#Object %s, %d polys, index %d:%d\n",
			part3d,
			((ObjectHeader*)object->object)->planes,
			*count,
			newCount++);
	*count = newCount;
	localOrigin[0] = origin[2] + flat->e3[0];
	localOrigin[1] = flat->e3[1];
	localOrigin[2] = origin[3] + flat->e3[2];
	localOrient[0] = flat->extents[0];
	localOrient[1] = flat->extents[1] + origin[4];
	localOrient[2] = flat->extents[2];
	gnuPlanes(object, part3d, localOrigin, localOrient, rectify);
	return 1;
}

char* printTownObjects
(	Query* query,
	char* handle,
	int count,
	int* index,
	long* origin,
	int rectify)
{
		/*	Describe a building in town, either its exterior view or its
			interior contents. */

	if (count)
		printf("#--  3D object: %d at 0x%x\n", count, handle - theData);
	while (count--)
	{
		if (!printTownObject(query, handle, index, origin, rectify))
			return NULL;
		handle = skip(Rmb3dObject);
	}
	return handle;
}

int printChunk
(	char* handle,
	RdbPartHeader* thread,
	Query* name,
	int* count)
{
		/*	Document the dependent portion of an rdb record. */

	Handle* object;
	char** slot;
	int auxData = 0;
	int override = 0;
	RdbPart1Data* ptr1;
	RdbPart2Data* ptr2;
	RdbPart3Data* ptr3;
	long next = thisPart + RdbPartHeaderSize;
	char part3d[8];
	char* objName;
	int newCount;

	switch ((int)thread->type)
	{
		case 2:
			next += sizeof(RdbPart2Data);
			/*ptr2 = (RdbPart2Data*)thread->data;
			printf(	"  %04lx %04lx type 2: %08lx"
					"                            %08lx %08lx %04hx   ",
					thisPart,
					next,
					ptr2->next,
					ptr2->data[0],
					ptr2->data[1],
					ptr2->code);*/
			break;
		case 0x81:
		case 1:
			next += sizeof(RdbPart1Data) - 10;
			ptr1 = (RdbPart1Data*)thread->data;
			if (name->filter && !matchObject(name, ptr1->objNo))
				break;
			/*auxData = ptr1->link > 0 && ptr1->link < 0xffff;*/
			override = ptr1->link == next;
			if (override)
				next += 10;
			objName = &handle[sizeof(RdbFixedHeader) + (ptr1->objNo<<3)];
			strncpy(part3d, objName, 5);
			part3d[5] = '\0';
			/*part3d[5] = '.';
			part3d[6] = '0';
			part3d[7] = '\0';*/
			object = *(Handle**)&handle[RdbOffsets + ((int)ptr1->objNo << 2)];
			if (!object)
			{
				object = open3dObject(name, part3d);
				if (!object)
				{
					printf("Can't open 3d object %s\n", part3d);
					exit(1);
				}
				*(char**)&handle[RdbOffsets + (ptr1->objNo << 2)] = (char*)object;
			}
			newCount = *count + ((ObjectHeader*)object->object)->planes - 1;
			printf(	"#Object %s, %d polys, index %d:%d\n",
					part3d,
					((ObjectHeader*)object->object)->planes,
					*count,
					newCount++);
			*count = newCount;
			printf(	"#  %04lx %04lx %04lx %04lx %08lx %08lx %08lx type %d%c %s\n",
					thisPart,
					next,
					thread->nextPart&0xffff,
					thread->priorPart&0xffff,
					thread->e3[0],
					thread->e3[1],
					thread->e3[2],
					(int)thread->type,
					' ',
					name->dungeon);
			printf(	"#  %04lx %04lx type 1: %04lx 0x%08lx 0x%08lx 0x%08lx"
					" %04hx %08lx %02x %08lx",
					thisPart,
					next,
					ptr1->next,
					ptr1->e3[0],
					ptr1->e3[1],
					ptr1->e3[2],
					ptr1->objNo,
					ptr1->data1,
					(int)ptr1->unknown1&0xff,
					ptr1->link);
			printf(" %s\n", name->dungeon);
			gnuPlanes(object, part3d, thread->e3, ptr1->e3, true);
			break;
		case 3:
			next += sizeof(RdbPart3Data);
			/*ptr3 = (RdbPart3Data*)thread->data;
			printf(	"  %04lx %04lx type 3: %08lx      %16.16s      %04lx"
					" %08x %08x %02x",
					thisPart,
					next,
					ptr3->next,
					findFlat(ptr3->picture),
					ptr3->picture,
					ptr3->data1,
					ptr3->link,
					(unsigned int)ptr3->data3&0xff);*/
			break;
		default:
			next += 3;
			return 0;
/*			printf("%d is an unknown chunk part type in", (int)thread->type);*/
			break;
	}
	/*if (auxData)
		printf(	" %s\n  %04lx %04lx type x: %08lx %02x %08lx %02x %s\n",
				name,
				ptr1->link,
				ptr1->link+10,
				ptr1->subdata,
				ptr1->subcode1,
				ptr1->sublink,
				ptr1->subcode2,
				name);
	else
		printf(" %s\n", name->dungeon);*/
	return override ? next : 0;
}

long nextChunk
(	char* handle,
	RdbPartHeader* current)
{
		/*	Given the current part header, read and return the
			next part header.  Answer nil if there is no next part. */

	int size;
	RdbPartHeader* privat = (RdbPartHeader*)&handle[thisPart];

	if (privat->nextPart < RdbTrailerOffset && privat->nextPart != -2 && privat->nextPart != -1)
		return thisPart + 10;
	if (privat->nextPart > 0xfffff)
		return thisPart + 10;
	size = privat->type & 0x7f;
	size = size > 3 ? 2 : (int)typeSizes[size] - 1;
	memcpy(current, privat, RdbPartHeaderSize + size);
	if ((current->type & 0x7f) == 1)
	{
		long pos = *(long*)&current->data[23];
		if (pos > 0 && pos < 0xffff)
		{
			memcpy(&current->data[27], &handle[pos], RdbPart1SubdataSize);
		}
	}
	return thisPart + size + RdbPartHeaderSize;
}

long printHeader
(	char* handle,
	RdbPartHeader* thread,
	Query* name)
{
		/*	Document the data type independent porition of an rdb record. */

	int die = 0;

	int auxData = 0;
	long nextPart;
	RdbPart1Data* ptr1;
	long next = thisPart + RdbPartHeaderSize;
	char extra = ' ';

	nextPart = nextChunk(handle, thread);
	if (thread->nextPart == 0xffffffff && nextPart < 1)
		return 0;
	switch ((int)thread->type)
	{
		case 2:
			next += sizeof(RdbPart2Data);
			break;
		case 0x81:
		case 1:
			next += sizeof(RdbPart1Data);
			ptr1 = (RdbPart1Data*)thread->data;
			auxData = ptr1->link > 0 && ptr1->link < 0xffff;
			next -= 10;
			if (auxData)
				extra = '*';
			auxData = 0;
			break;
		case 3:
			next += sizeof(RdbPart3Data);
			break;
		default:
		next += 3;
		/*if (nextPart)
			printf(	"  %04x Skip to next part %08x\n",
					thisPart,
					nextPart);*/
		return nextPart;
/*		printf(	"%d is an unknown part at %x type in",
				(int)thread->type,
				thread->nextPart);
		die = 1;*/
		break;
	}
	/*printf(	"  %04x %04x %04x %04x %08x %08x %08x type %d%c %s\n",
			thisPart,
			next,
			thread->nextPart&0xffff,
			thread->priorPart&0xffff,
			thread->e3[0],
			thread->e3[1],
			thread->e3[2],
			(int)thread->type,
			extra,
			name->dungeon);*/
	if (die)
		nextPart = -1;
	return nextPart;
}

int printChunks
(	char* handle,
	RdbFixedHeader* fixed,
	long fileSize,
	Query* name)
{
		/*	Digest the variable part of an rdb file in strict serial
			order. */

	RdbPartHeader thread;
	int die = 0;
	int count = 0;

	thisPart = fixed->bodyStart+64;
	memcpy(&thread, &handle[thisPart], sizeof(RdbPartHeader));
	while (thisPart < fileSize)
	{
		long nextPart;
		long override;

		nextPart = printHeader(handle, &thread, name);
		/*if (thread.nextPart == 0xffffffff || nextPart == 0xffffffff)
			break;*/
		override = printChunk(handle, &thread, name, &count);
		thisPart = override? override : nextPart;
	}
	return 1;
}

RdbPartHeader* nextPart
(	char* handle,
	RdbPartHeader* current)
{
		/*	Given the current part header, read and return the
			next part header.  Answer nil if there is no next part. */

	int size;
	RdbPartHeader* privat;

	if (current->nextPart <= 0)
		return NULL;
	thisPart = current->nextPart;
	privat = (RdbPartHeader*)&handle[thisPart];
	size = privat->type & 0x7f;
	size = (int)typeSizes[size] - 1;
	memcpy(current, privat, RdbPartHeaderSize + size);
	if ((current->type & 0x7f) == 1)
	{
		long pos = *(long*)&current->data[23];
		if (pos > 0 && pos < 0xffff)
		{
			memcpy(&current->data[27], &handle[pos], RdbPart1SubdataSize);
		}
	}
	return current;
}

int updateList
(	unsigned short* list,
	long offset)
{
		/*	Mark off an rdb record appearing in a linked list. */

	unsigned short off = offset;
	int count = elements;

	while (count--)
	{
		if (!*list)
		{
			list = &list[1];
			continue;
		}
		if (*list == off)
		{
			*list = 0;
			return 1;
		}
		if (*list > off)
			return 0;
		list = &list[1];
	}
	return 0;
}

int printThread
(	char* handle,
	RdbPartHeader* thread,
	Query* name,
	int* count)
{
		/*	Digest a canonical linked list in an rdb file. */

	while (nextPart(handle, thread))
	{
		printChunk(handle, thread, name, count);
	}
	return 1;
}

int countObjects
(	char* handle)
{
		/*	Answer the count of 3d objects used in the file. */

	Rdb3dObjectDir* objDir;

	trailer = (RdbFixedTrailer*)&handle[RdbTrailerOffset];
	printf(	"#Fixed trailer: %08lx %08lx\n",
			trailer->unknown4,
			trailer->bodyEnd);
	objDir = (Rdb3dObjectDir*)&handle[trailer->to3dObjectDir];
	return 1+(int)objDir->itemNo;
}

int print3dObjects
(	Query* query,
	char* handle,
	int oCount)
{
		/*	Print the requested number of object profiles. */

	char* objName;
	long objRec = trailer->to3dObjectDir;
	int count = 0;
	int newCount = 0;
	long e3[3] = {0, 0, 0};

	objName = (char*)&handle[sizeof(RdbFixedHeader)];
	printf("#3D Objects: %d\n", oCount);
	while (oCount--)
	{
		char part[4];
		char part3d[8];
		Rdb3dObjectDir* objDir;
		Handle* object;

		strncpy(part3d, objName, 5);
		part3d[5] = '\0';
		strncpy(part, &objName[5], 3);
		part[3] = '\0';
		objDir = (Rdb3dObjectDir*)&handle[objRec];
		printf(	"#  %04lx %04lx   %s %s %4hd %08lx %08lx %08lx %08lx\n",
				objRec,
				objDir->next&0xffff,
				part3d,
				part,
				objDir->itemNo,
				objDir->nextPart,
				objDir->part[0],
				objDir->part[1],
				objDir->part[2]
				);
		part3d[5] = '\0';
		/*part3d[5] = '.';
		part3d[6] = '0';
		part3d[7] = '\0';*/
		object = *(Handle**)&handle[RdbOffsets + ((int)objDir->itemNo << 2)];
		if (!object)
		{
			object = open3dObject(query, part3d);
			if (!object)
			{
				printf("Can't open 3d object %s\n", part3d);
				exit(1);
			}
			*(char**)&handle[RdbOffsets + (objDir->itemNo << 2)] = (char*)object;
		}
		newCount = count + ((ObjectHeader*)object->object)->planes - 1;
		printf(	"#Object %s, %d polys, index %d:%d\n",
				part3d,
				((ObjectHeader*)object->object)->planes,
				count,
				newCount++);
		count = newCount;
		gnuPlanes(object, part3d, e3, e3, true);
		objRec = objDir->next;
		objName += 8;
	}
	return 1;
}

void findUnlinked
(	unsigned short* list,
	char* name)
{
		/*	Report any unlinked elements. */

	int count = elements;

	printf("\n");
	while (count--)
	{
		if (*list)
			printf(	"  %04hx unlinked in %s.\n",
					*list,
					name);
		list++;
	}
}

Query* asPartList
(	char* list)
{
		/*	Digest the command list of requested object part numbers. */

	char* punt = strchr(list, '.');
	theQuery.dungeon = list;

	if (theQuery.partList)
		free(theQuery.partList);
	theQuery.partList = NULL;
	theQuery.partCount = 0;
	theQuery.filter = punt[4] == ':';
	if (theQuery.filter)
	{
		short* parts;
		int partCount;

		list = &punt[4];
		partCount = 1 + (strlen(&punt[4]) >> 2);
		parts = theQuery.partList = (short*)malloc(partCount << 1);
		while (*list == ':')
		{
			char* next;

			*list++ = '\0';
			next = strchr(list, ':');
			printf("#-- Building subpart list.\n");
			if (!next)
				next = &list[strlen(list)];
			*parts++ = (short)(list - theQuery.dungeon);
			theQuery.partCount++;
			list = next;
		}
	}
	return &theQuery;
}

int sortParts
(	const void* e1,
	const void* e2)
{
		/*	qsort() appendage to sort part indices. */

	return (int)(*(short*)e1 - *(short*)e2);
}

int asObjNumbers
(	char* handle,
	Query* query)
{
		/*	Convert the list of part numbers into a sorted index of
			part numbers. */

	char* objName;
	short* newParts;
	short* partList;

	if (!query->filter)
		return 1;
	if (2 == query->rmbs)
		objName = "OOOOO";
	else
		objName = &handle[sizeof(RdbFixedHeader)];
	newParts = partList = (short*)malloc(query->partCount << 1);
	memset(partList, 0, query->partCount << 1);
	while (*objName > 0)
	{
		int count = query->partCount;
		short* parts = query->partList;
		char oNam[6];

		strncpy(oNam, objName, 5);
		oNam[5] = '\0';
		while (count--)
		{
			if (*parts < 0)
			{
				parts++;
				continue;
			}
			if (2 == query->rmbs)
			{
				/*	Translate a building number filter. */

				*newParts++ = (short)atoi(&query->dungeon[*parts]);
				*parts++ = -1;
				continue;
			}
			if (!strncmp(&query->dungeon[*parts++], objName, 5))
			{
				*newParts++ = (objName - &handle[sizeof(RdbFixedHeader)]) >> 3;
				break;
			}
		}
		if (2 == query->rmbs)
			break;
		objName += 8;
	}
	free(query->partList);
	qsort(partList, query->partCount, 2, sortParts);
	query->partList = partList;
	return 1;
}

int describeDungeon
(	int useThreads,
	char** argv,
	Query* query)
{
		/*	Describe a dungeon scenario, filtered according to cli. */

	char* data;
	int count;
	long fileSize;
	long final;
	RdbFixedHeader* fixed;
	RdbPartAnchor* anchors;
	RdbPartHeader thread;
	RdbPartAnchor* anch;
	int index;
	unsigned short* list;
	unsigned short* lPtr, *pPtr;
	int theCount = 0;

	/*	Locate the requested scenario. */

	final = fileSize = openScenario(query);
	if (!fileSize)
		return 0;
	(*argv)[8] = '\0';
	data = (char*)malloc(fileSize);
	fixed = (RdbFixedHeader*)data;
	read(query->handle, data, fileSize);
	memset(&data[RdbOffsets], 0, RdbOffsetSize);

	/*	Finalize the part number search sequence and walk the chains
		of linked lists in the scenario. */

	asObjNumbers(data, query);
	anch = anchors = (RdbPartAnchor*)&data[fixed->bodyStart];
	if (useThreads == 1)
	{
		count = 4;
		while (count--)
		{
			/*	Walk a thread. */

			long* threads = anch++->threads;
			int tCount = 1;

			index = 4;
			while (index--)
			{
				thread.nextPart = *threads++;
				if (thread.nextPart < 0)
					continue;
				printf("#-- Anchor %d, thread %d (%x).\n", index, tCount++, thread.nextPart);
				printThread(data, &thread, query, &theCount);
			}
		}
	}
	else if (useThreads == 2)
		print3dObjects(query, data, countObjects(data));
	else
		printChunks(data, fixed, fileSize, query);
	free(data);
	return 1;
}

int describeTown
(	char** argv,
	Query* query)
{
		/*	Describe a town scenario, filtered according to the cli. */

	RmbHeader* header;
	RmbFixedHeader* fixed;
	int handle = -1;
	int count;
	long* size;
	long fileSize;
	char* data;
	char* current;
	long* sizes;
	long position;
	long* origin;
	int index = 0;
	int building = 0;
	long cellOrigin[] = {0, 0, 0, 4096, 0};

	fileSize = openScenario(query);
	if (!fileSize)
	{
		perror("open");
		fprintf(stderr, "Can't open %s.\n", query->dungeon);
		exit(1);
	}
	theData = data = (char*)malloc(fileSize);
	read(query->handle, data, fileSize);
	fixed = (RmbFixedHeader*)data;
	asObjNumbers(data, query);
	printf(	"#-- Fixed: blocks %2d, extra objects %2d, extra flats %2d, %s\n",
			fixed->blocks,
			fixed->objs,
			fixed->flats,
			*argv);
	current = &data[sizeof(RmbFixedHeader)];
	origin = (long*)current;
	/*origin = &origin[2];*/
	/*current = print2(Positions, current, fixed);
	current = print2(DoorStops, current, fixed);
	current = print2(Offsets, current, fixed);*/
	current = &data[RmbSizes];
	size = sizes = (long*)current;
/*	print2(Sizes, handle, fixed);*/
/*	printFloorMap(handle, *argv);*/
	current = &data[RmbFixedSize];

	/*	Iterate over the number of embedded buildings. */

	count = fixed->blocks;
	while (count--)
	{
		/*	Handle the outer block records. */

		building++;
		position = current - data;
		printf("#-- Begin block %d at 0x%x\n", count, position);
		if (position + sizeof(RmbHeader) > fileSize)
		{
			break;
		}
		if (query->filter && !matchObject(query, building))
		{
			position += *size++;
			current = &data[position];
			origin = &origin[5];
			continue;
		}
		origin[3] = 4096 - origin[3];
		header = (RmbHeader*)current;
		current = &current[sizeof(RmbHeader)];
		printf(	"#-- Outers: %2d %2d %2d %2d %2d 0x%08x 0x%08x 0x%08x\n",
				(int)header->objs,
				(int)header->flats,
				(int)header->sections,
				(int)header->npcs,
				(int)header->doors,
				header->links[0],
				header->links[1],
				header->links[2]);
		current = printTownObjects(query, current, (int)header->objs, &index, origin, true);
		/*current = print2(Flats, current, header);
		current = print3(Sections, current, header, fixed);
		current = print2(Npcs, current, header);
		current = print2(Doors, current, header);*/
		if (1 == query->rmbs)
		{
			current +=	(header->flats * sizeof(RmbFlat)) +
						(header->sections * sizeof(RmbSection)) +
						(header->npcs * sizeof(RmbNpc)) +
						(header->doors * sizeof(RmbDoor));
		}
		else
		{
			/*	Handle the inner block records. */

		/*	current += sizeof(Rmb3dObject) * header->objs;*/
			current += sizeof(RmbFlat) * header->flats;
			current += sizeof(RmbSection) * header->sections;
			current += sizeof(RmbNpc) * header->npcs;
			current += sizeof(RmbDoor) * header->doors;
			if ((current - data) + sizeof(RmbHeader) > fileSize)
			{
				break;
			}
			header = (RmbHeader*)current;
			printf(	"#-- Inners: %2d %2d %2d %2d %2d 0x%08x 0x%08x 0x%08x\n",
					(int)header->objs,
					(int)header->flats,
					(int)header->sections,
					(int)header->npcs,
					(int)header->doors,
					header->links[0],
					header->links[1],
					header->links[2]);
			printf("#-- Begin inners at 0x%x\n", current - data);
			current = &current[sizeof(RmbHeader)];
			current = printTownObjects(query, current, (int)header->objs, &index, origin, false);
		}
/*		current = print2(Flats, current, header);
		current = print3(Sections, current, header, fixed);
		current = print2(Npcs, current, header);
		current = print2(Doors, current, header);*/
		position += *size++;
		current = &data[position];
		origin = &origin[5];
	}
	/*	Handle any trailing arch3d entries. */

	printf("#-- Extra 3D Objects: %d\n", count);
	current = printTownObjects(query, current, (int)fixed->objs, &index, cellOrigin, true);

	/*	Handle any trailing flats. */

	current = printFlatLabels(current, (int)fixed->flats, &cellOrigin[1]);
/*	count = fixed->flats;
	printf("-- Extra flats: %d\n", count);
	while (count--)
	{
		printFlat(current);
		current = &current[sizeof(RmbFlat)];
		position = current - data;
		if (position > fileSize)
			break;
	}*/
/*	printf("-- Last offset: 0x%x\n", lseek(handle, 0L, SEEK_CUR));*/
	free(data);
	return 1;
}

int main
(	int argc,
	char** argv)
{
	char thePath[256];
	char* path;
	char obj[256];
	int useThreads = 1;
	int o3dHandle;
	int handle = -1;
	int objCount;
	long* objDir;

/*	pSize(RdbFixedHeader);
	pSize(RdbFixedTrailer);
	pSize(Rdb3dObjectDir);
	pSize(RdbPartAnchor);
	pSize(RdbPartHeader);
	pSize(RdbPosition);
	pSize(RdbPart1Data);
	pSize(RdbPart2Data);
	pSize(RdbPart3Data);*/

	if (argc == 1)
	{
		printf( "rooms:  Daggerfall Polygon Extractor, version 1.0\n"
				"        Copyright (c) 2002, Nagual Enterprises\n\n"
				"Usage:  rooms  [-io] room1  room2 ...\n"
				"        where roomN is the file name of an rdb/rmb file in blocks.bsa.\n"
				"        -o selects only the dungeon object directory; no repeated objects.\n"
				"        -i selects only the town interiors.  (By default, only\n"
				"        the town exteriors are selected.  You don't want both)\n"
				"        roomN can contain a colon separated list of objects to extract.\n"
				"        (By default all objects are included)\n\n"
				"          rooms s0000000.rdb:74039:74080\n\n"
				"        extracts all the 74039 and 74080 objects from the first\n"
				"        chamber of the Mantellan Crux.\n\n"
				"          rooms -i mageaa08.rmb:1\n\n"
				"        selects the interior of the Mage's guild in Daggerfall City.\n"
				"        Set TES2 to the Daggerfall installation to use for extraction.\n");
		return 0;
	}
	mTwoPi = -twoPi;
	o3d = getenv("TES2");
	if (o3d)
	{
		strcpy(stpcpy(obj, o3d), DirSep "arena2" DirSep "arch3d.bsa");
		strcpy(stpcpy(thePath, o3d), DirSep "arena2" DirSep "blocks.bsa");
		flats = stpcpy(stpcpy(theFlats, o3d), DirSep);
	}
	else
	{
		strcpy(obj, "." DirSep "arena2" DirSep "arch3d.bsa");
		strcpy(thePath, "." DirSep "arena2" DirSep "blocks.bsa");
		flats = stpcpy(theFlats, "." DirSep);
	}
	o3d = obj;
	o3dHandle = open(o3d, O_RDONLY | O_BINARY);
	if (o3dHandle < 0)
	{
		perror("open");
		fprintf(stderr, "Can't open %s.\n", o3d);
		exit(1);
	}
	/*	Build an inventory of the object repository. */

	objDir = buildObjDir(o3dHandle, o3d, &objCount);
	handle = open(thePath, O_RDONLY | O_BINARY);
	if (handle < 0)
	{
		perror("open");
		fprintf(stderr, "Can't open %s.\n", thePath);
		exit(1);
	}
	theQuery.rmbs = 1;
	if (argv[1][0] == '-')
	{
		if ('o' == argv[1][1])
			useThreads = 2;
		else if ('i' == argv[1][1])
			theQuery.rmbs = 2;
		argv++;
	}
	theQuery.o3dHandle = o3dHandle;
	theQuery.handle = handle;
	theQuery.objDir = objDir;
	theQuery.objCount = objCount;
	while (*++argv)
	{
		Query* query;
		char* punt;

		/*	Build the request for this scenario. */

		punt = strchr(*argv, '.');
		if (!punt)
		{
			printf("%s is not a town or dungeon scenario.\n", *argv);
			continue;
		}
		query = asPartList(*argv);

		if (isEqualI(punt, ".rdb", 4))
			describeDungeon(useThreads, argv, query);
		else if (isEqualI(punt, ".rmb", 4))
			describeTown(argv, query);
		else
			printf("%s is not a town or dungeon scenario.\n", *argv);
	}
	return 0;
}

